"""
Spout Specification

This module is called spout to mirror organization of storm package.
"""

from .component import Specification
from ..storm.spout import Spout


class SpoutSpecification(Specification):
    def __init__(self, component_cls, **kwargs):
        if not issubclass(component_cls, Spout):
            raise TypeError("Invalid spout: {}".format(component_cls))

        return super(SpoutSpecification, self).__init__(component_cls, **kwargs)



"""
Run a local Storm topology.
"""

from __future__ import absolute_import, print_function

import os
import sys

from invoke import run

from .common import (add_ackers, add_debug, add_environment, add_name,
                     add_options, add_par, add_workers, resolve_ackers_workers)
from ..util import prepare_topology, get_topology_definition


def run_local_topology(name=None, time=0, workers=2, ackers=2, options=None,
                       debug=False):
    """Run a topology locally using Storm's LocalCluster class."""
    prepare_topology()

    name, topology_file = get_topology_definition(name)
    print("Running {} topology...".format(name))
    sys.stdout.flush()
    cmd = ["lein",
           "run -m streamparse.commands.run/-main",
           topology_file]
    cmd.append("-t {}".format(time))
    if debug:
        cmd.append("--debug")
    cmd.append("--option 'topology.workers={}'".format(workers))
    cmd.append("--option 'topology.acker.executors={}'".format(ackers))

    # Python logging settings
    if not os.path.isdir("logs"):
        os.makedirs("logs")
    log_path = os.path.join(os.getcwd(), "logs")
    print("Routing Python logging to {}.".format(log_path))
    sys.stdout.flush()
    cmd.append("--option 'streamparse.log.path=\"{}\"'"
               .format(log_path))
    cmd.append("--option 'streamparse.log.level=\"debug\"'")

    if options is None:
        options = []
    for option in options:
        cmd.append('--option {}'.format(option))
    full_cmd = " ".join(cmd)
    print("Running lein command to run local cluster:")
    print(full_cmd)
    sys.stdout.flush()
    run(full_cmd)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('run',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_ackers(subparser)
    add_debug(subparser)
    add_environment(subparser)
    add_name(subparser)
    add_options(subparser)
    add_par(subparser)
    subparser.add_argument('-t', '--time',
                           default=0,
                           type=int,
                           help='Time (in seconds) to keep local cluster '
                                'running. If time <= 0, run indefinitely. '
                                '(default: %(default)s)')
    add_workers(subparser)


def main(args):
    """ Run the local topology with the given arguments """
    resolve_ackers_workers(args)
    run_local_topology(name=args.name, time=args.time, workers=args.workers,
                       ackers=args.ackers, options=args.options,
                       debug=args.debug)



"""
Tests for Bolt and BatchingBolt classes
"""

from __future__ import absolute_import, print_function, unicode_literals

import logging
import unittest
from io import BytesIO

import simplejson as json

try:
    from unittest import mock
    from unittest.mock import patch
except ImportError:
    import mock
    from mock import patch

from streamparse.storm import BatchingBolt, Bolt, Tuple


log = logging.getLogger(__name__)


class BoltTests(unittest.TestCase):

    def setUp(self):
        self.tup_dict = {'id': 14,
                         'comp': 'some_spout',
                         'stream': 'default',
                         'task': 'some_bolt',
                         'tuple': [1, 2, 3]}
        tup_json = "{}\nend\n".format(json.dumps(self.tup_dict)).encode('utf-8')
        self.tup = Tuple(self.tup_dict['id'], self.tup_dict['comp'],
                         self.tup_dict['stream'], self.tup_dict['task'],
                         self.tup_dict['tuple'],)
        self.bolt = Bolt(input_stream=BytesIO(tup_json),
                         output_stream=BytesIO())
        self.bolt.initialize({}, {})

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_emit_basic(self, send_message_mock):
        # A basic emit
        self.bolt.emit([1, 2, 3], need_task_ids=False)
        send_message_mock.assert_called_with(self.bolt, {'command': 'emit',
                                                         'anchors': [],
                                                         'tuple': [1, 2, 3],
                                                         'need_task_ids': False})

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_emit_stream_anchors(self, send_message_mock):
        # Emit with stream and anchors
        self.bolt.emit([1, 2, 3], stream='foo', anchors=[4, 5],
                       need_task_ids=False)
        send_message_mock.assert_called_with(self.bolt, {'command': 'emit',
                                                         'stream': 'foo',
                                                         'anchors': [4, 5],
                                                         'tuple': [1, 2, 3],
                                                         'need_task_ids': False})

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_emit_direct(self, send_message_mock):
        # Emit as a direct task
        self.bolt.emit([1, 2, 3], direct_task='other_bolt')
        send_message_mock.assert_called_with(self.bolt, {'command': 'emit',
                                                         'anchors': [],
                                                         'tuple': [1, 2, 3],
                                                         'task': 'other_bolt'})

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_emit_many_basic(self, send_message_mock):
        # A basic emit
        self.bolt.emit_many([[1, 2, 3], [4, 5, 6]], need_task_ids=False)
        send_message_mock.assert_has_calls([mock.call(self.bolt,
                                                      {'command': 'emit',
                                                       'tuple': [1, 2, 3],
                                                       'anchors': [],
                                                       'need_task_ids': False}),
                                            mock.call(self.bolt,
                                                      {'command': 'emit',
                                                       'tuple': [4, 5, 6],
                                                       'anchors': [],
                                                       'need_task_ids': False})])

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_emit_many_anchors(self, send_message_mock):
        # Emit with anchors
        self.bolt.emit_many([[1, 2, 3], [4, 5, 6]], anchors=['foo', 'bar'],
                             need_task_ids=False)
        send_message_mock.assert_has_calls([mock.call(self.bolt,
                                                      {'command': 'emit',
                                                       'tuple': [1, 2, 3],
                                                       'need_task_ids': False,
                                                       'anchors': ['foo', 'bar']}),
                                            mock.call(self.bolt,
                                                      {'command': 'emit',
                                                       'tuple': [4, 5, 6],
                                                       'anchors': ['foo', 'bar'],
                                                       'need_task_ids': False})])

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_emit_many_direct(self, send_message_mock):
        # Emit as a direct task
        self.bolt.emit_many([[1, 2, 3], [4, 5, 6]], direct_task='other_bolt')
        send_message_mock.assert_has_calls([mock.call(self.bolt,
                                                      {'command': 'emit',
                                                       'anchors': [],
                                                       'tuple': [1, 2, 3],
                                                       'task': 'other_bolt'}),
                                            mock.call(self.bolt,
                                                      {'command': 'emit',
                                                       'anchors': [],
                                                       'tuple': [4, 5, 6],
                                                       'task': 'other_bolt'})])

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_ack_id(self, send_message_mock):
        # ack an ID
        self.bolt.ack(42)
        send_message_mock.assert_called_with(self.bolt, {'command': 'ack',
                                                         'id': 42})

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_ack_tuple(self, send_message_mock):
        # ack a Tuple
        self.bolt.ack(self.tup)
        send_message_mock.assert_called_with(self.bolt, {'command': 'ack',
                                                         'id': 14})

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_fail_id(self, send_message_mock):
        # fail an ID
        self.bolt.fail(42)
        send_message_mock.assert_called_with(self.bolt, {'command': 'fail',
                                                         'id': 42})

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_fail_tuple(self, send_message_mock):
        # fail a Tuple
        self.bolt.ack(self.tup)
        send_message_mock.assert_called_with(self.bolt, {'command': 'ack',
                                                         'id': 14})

    @patch.object(Bolt, 'process', autospec=True)
    @patch.object(Bolt, 'ack', autospec=True)
    def test_run(self, ack_mock, process_mock):
        self.bolt._run()
        process_mock.assert_called_with(self.bolt, self.tup)
        self.assertListEqual(self.bolt._current_tups, [])


    @patch.object(Bolt, 'process', autospec=True)
    @patch.object(Bolt, 'ack', autospec=True)
    def test_auto_ack_on(self, ack_mock, process_mock):
        # test auto-ack on (the default)
        self.bolt._run()
        ack_mock.assert_called_with(self.bolt, self.tup)
        self.assertEqual(ack_mock.call_count, 1)

    @patch.object(Bolt, 'process', autospec=True)
    @patch.object(Bolt, 'ack', autospec=True)
    def test_auto_ack_off(self, ack_mock, process_mock):
        self.bolt.auto_ack = False
        self.bolt._run()
        # Assert that this wasn't called, and print out what it was called with
        # otherwise.
        self.assertListEqual(ack_mock.call_args_list, [])

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_auto_anchor_on(self, send_message_mock):
        self.bolt._current_tups = [self.tup]
        # Test auto-anchor on (the default)
        self.bolt.emit([1, 2, 3], need_task_ids=False)
        send_message_mock.assert_called_with(self.bolt, {'command': 'emit',
                                                         'anchors': [14],
                                                         'tuple': [1, 2, 3],
                                                         'need_task_ids': False})

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_auto_anchor_off(self, send_message_mock):
        # Test auto-anchor off
        self.bolt.auto_anchor = False
        self.bolt.emit([1, 2, 3], need_task_ids=False)
        send_message_mock.assert_called_with(self.bolt, {'command': 'emit',
                                                         'anchors': [],
                                                         'tuple': [1, 2, 3],
                                                         'need_task_ids': False})

    @patch.object(Bolt, 'send_message', autospec=True)
    def test_auto_anchor_override(self, send_message_mock):
        # Test overriding auto-anchor
        self.bolt.auto_anchor = True
        self.bolt.emit([1, 2, 3], anchors=[42], need_task_ids=False)
        send_message_mock.assert_called_with(self.bolt, {'command': 'emit',
                                                         'anchors': [42],
                                                         'tuple': [1, 2, 3],
                                                         'need_task_ids': False})

    @patch('sys.exit', new=lambda r: r)
    @patch.object(Bolt, 'read_handshake', new=lambda x: ({}, {}))
    @patch.object(Bolt, 'fail', autospec=True)
    @patch.object(Bolt, '_run', autospec=True)
    def test_auto_fail_on(self, _run_mock, fail_mock):
        self.bolt._current_tups = [self.tup]
        # Make sure _run raises an exception
        def raiser(): # lambdas can't raise
            raise Exception('borkt')
        _run_mock.side_effect = raiser

        # test auto-fail on (the default)
        self.bolt.run()
        fail_mock.assert_called_with(self.bolt, self.tup)
        self.assertEqual(fail_mock.call_count, 1)

    @patch('sys.exit', new=lambda r: r)
    @patch.object(Bolt, 'read_handshake', new=lambda x: ({}, {}))
    @patch.object(Bolt, 'raise_exception', new=lambda *a: None)
    @patch.object(Bolt, 'fail', autospec=True)
    @patch.object(Bolt, '_run', autospec=True)
    def test_auto_fail_off(self, _run_mock, fail_mock):
        self.bolt._current_tups = [self.tup]
        # Make sure _run raises an exception
        def raiser(): # lambdas can't raise
            raise Exception('borkt')
        _run_mock.side_effect = raiser

        # test auto-fail off
        self.bolt.auto_fail = False
        self.bolt.run()
        # Assert that this wasn't called, and print out what it was called with
        # otherwise.
        self.assertListEqual(fail_mock.call_args_list, [])

    @patch.object(Bolt, 'read_tuple', autospec=True)
    @patch.object(Bolt, 'send_message', autospec=True)
    def test_heartbeat_response(self, send_message_mock, read_tuple_mock):
        # Make sure we send sync for heartbeats
        read_tuple_mock.return_value = Tuple(id='foo', task=-1,
                                             stream='__heartbeat', values=[],
                                             component='__system')
        self.bolt._run()
        send_message_mock.assert_called_with(self.bolt, {'command': 'sync'})

    @patch.object(Bolt, 'read_tuple', autospec=True)
    @patch.object(Bolt, 'process_tick', autospec=True)
    def test_process_tick(self, process_tick_mock, read_tuple_mock):
        # Make sure we send sync for heartbeats
        read_tuple_mock.return_value = Tuple(id=None, task=-1,
                                             component='__system',
                                             stream='__tick', values=[50])
        self.bolt._run()
        process_tick_mock.assert_called_with(self.bolt,
                                             read_tuple_mock.return_value)


class BatchingBoltTests(unittest.TestCase):

    def setUp(self):
        self.ticks_between_batches = 1
        self.tup_dicts = [{'id': 14,
                           'comp': 'some_spout',
                           'stream': 'default',
                           'task': 'some_bolt',
                           'tuple': [1, 2, 3]},
                          {'id': 15,
                           'comp': 'some_spout',
                           'stream': 'default',
                           'task': 'some_bolt',
                           'tuple': [4, 5, 6]},
                          {'id': None,
                           'comp': '__system',
                           'stream': '__tick',
                           'task': -1,
                           'tuple': [1]},
                          {'id': 16,
                           'comp': 'some_spout',
                           'stream': 'default',
                           'task': 'some_bolt',
                           'tuple': [7, 8, 9]},
                          {'id': None,
                           'comp': '__system',
                           'stream': '__tick',
                           'task': -1,
                           'tuple': [2]}]
        tups_json = '\nend\n'.join([json.dumps(tup_dict) for tup_dict in
                                    self.tup_dicts] + [''])
        self.tups = [Tuple(tup_dict['id'], tup_dict['comp'], tup_dict['stream'],
                           tup_dict['task'], tup_dict['tuple']) for tup_dict in
                     self.tup_dicts]
        self.nontick_tups = [tup for tup in self.tups if tup.stream != '__tick']
        self.bolt = BatchingBolt(input_stream=BytesIO(tups_json.encode('utf-8')),
                                 output_stream=BytesIO())
        self.bolt.initialize({}, {})

    @patch.object(BatchingBolt, 'process_batch', autospec=True)
    def test_batching(self, process_batch_mock):
        # Add a bunch of Tuples
        for __ in self.tups:
            self.bolt._run()

        process_batch_mock.assert_called_with(self.bolt, None,
                                              self.nontick_tups)

    @patch.object(BatchingBolt, 'process_batch', autospec=True)
    def test_group_key(self, process_batch_mock):
        # Change the group key to even/odd grouping
        self.bolt.group_key = lambda t: sum(t.values) % 2

        # Add a bunch of Tuples
        for __ in self.tups:
            self.bolt._run()

        process_batch_mock.assert_has_calls([mock.call(self.bolt, 0,
                                                       [self.nontick_tups[0],
                                                        self.nontick_tups[2]]),
                                             mock.call(self.bolt, 1,
                                                       [self.nontick_tups[1]])],
                                            any_order=True)

    @patch.object(BatchingBolt, 'ack', autospec=True)
    @patch.object(BatchingBolt, 'process_batch', new=lambda *args: None)
    def test_auto_ack_on(self, ack_mock):
        # Test auto-ack on (the default)
        for __ in self.tups:
            self.bolt._run()
        ack_mock.assert_has_calls([mock.call(self.bolt, tup)
                                   for tup in self.tups],
                                  any_order=True)
        self.assertEqual(ack_mock.call_count, 5)

    @patch.object(BatchingBolt, 'ack', autospec=True)
    @patch.object(BatchingBolt, 'process_batch', new=lambda *args: None)
    def test_auto_ack_off(self, ack_mock):
        # Test auto-ack off
        self.bolt.auto_ack = False
        for __ in self.tups:
            self.bolt._run()
        # Assert that this wasn't called, and print out what it was called with
        # otherwise.
        ack_mock.assert_has_calls([mock.call(self.bolt, tup)
                                   for tup in self.tups
                                   if self.bolt.is_tick(tup)],
                                  any_order=True)
        self.assertEqual(ack_mock.call_count, 2)

    @patch.object(BatchingBolt, 'read_handshake', new=lambda x: ({}, {}))
    @patch.object(BatchingBolt, 'raise_exception', new=lambda *a: None)
    @patch('sys.exit', autospec=True)
    @patch.object(BatchingBolt, 'fail', autospec=True)
    def test_auto_fail_on(self, fail_mock, exit_mock):
        # Need to re-register signal handler with mocked version, because
        # mock gets created after handler was originally registered.
        self.setUp()
        # Test auto-fail on (the default)
        self.bolt.run()

        # All waiting Tuples should have failed at this point
        fail_mock.assert_has_calls([mock.call(self.bolt, self.nontick_tups[0]),
                                    mock.call(self.bolt, self.nontick_tups[1]),
                                    mock.call(self.bolt, self.nontick_tups[2])],
                                   any_order=True)
        self.assertEqual(fail_mock.call_count, 3)
        self.assertEqual(exit_mock.call_count, 1)

    @patch.object(BatchingBolt, 'read_handshake', new=lambda x: ({}, {}))
    @patch.object(BatchingBolt, 'raise_exception', new=lambda *a: None)
    @patch('sys.exit', autospec=True)
    @patch.object(BatchingBolt, 'fail', autospec=True)
    def test_auto_fail_off(self, fail_mock, exit_mock):
        # Need to re-register signal handler with mocked version, because
        # mock gets created after handler was originally registered.
        self.setUp()
        # Test auto-fail off
        self.bolt.auto_fail = False
        self.bolt.run()

        # All waiting Tuples should have failed at this point
        self.assertEqual(exit_mock.call_count, 1)
        self.assertListEqual(fail_mock.call_args_list, [])

    @patch.object(BatchingBolt, 'read_handshake', new=lambda x: ({}, {}))
    @patch('sys.exit', autospec=True)
    @patch.object(BatchingBolt, 'process_batch', autospec=True)
    @patch.object(BatchingBolt, 'fail', autospec=True)
    def test_auto_fail_partial(self, fail_mock, process_batch_mock, exit_mock):
        # Need to re-register signal handler with mocked version, because
        # mock gets created after handler was originally registered.
        self.setUp()
        # Change the group key just be the sum of values, which makes 3 separate
        # batches
        self.bolt.group_key = lambda t: sum(t.values)
        # Make sure we fail on the second batch
        work = {'status': True} # to avoid scoping problems
        def work_once(*args):
            if work['status']:
                work['status'] = False
            else:
                raise Exception('borkt')
        process_batch_mock.side_effect = work_once
        # Run the batches
        self.bolt.run()
        # Only some Tuples should have failed at this point. The key is that
        # all un-acked Tuples should be failed, even for batches we haven't
        # started processing yet.
        self.assertEqual(fail_mock.call_count, 2)
        self.assertEqual(exit_mock.call_count, 1)

    @patch.object(BatchingBolt, 'read_tuple', autospec=True)
    @patch.object(BatchingBolt, 'send_message', autospec=True)
    def test_heartbeat_response(self, send_message_mock, read_tuple_mock):
        # Make sure we send sync for heartbeats
        read_tuple_mock.return_value = Tuple(id='foo', task=-1,
                                             stream='__heartbeat', values=[],
                                             component='__system')
        self.bolt._run()
        send_message_mock.assert_called_with(self.bolt, {'command': 'sync'})

    @patch.object(BatchingBolt, 'read_tuple', autospec=True)
    @patch.object(BatchingBolt, 'process_tick', autospec=True)
    def test_process_tick(self, process_tick_mock, read_tuple_mock):
        # Make sure we send sync for heartbeats
        read_tuple_mock.return_value = Tuple(id=None, task=-1,
                                             component='__system',
                                             stream='__tick', values=[50])
        self.bolt._run()
        process_tick_mock.assert_called_with(self.bolt,
                                             read_tuple_mock.return_value)



if __name__ == '__main__':
    unittest.main()



"""
Bolt Specification

This module is called bolt to mirror organization of storm package.
"""

from collections import Iterable
from six import string_types

from .component import Specification
from .topology import Grouping, TopologyError
from ..storm.bolt import Bolt


class BoltSpecification(Specification):
    def __init__(self, component_cls, source=None, group_on=Grouping.SHUFFLE,
                 **kwargs):
        if not issubclass(component_cls, Bolt):
            raise TypeError("Invalid bolt: {}".format(component_cls))

        super(BoltSpecification, self).__init__(component_cls, **kwargs)

        # Ensure that the source is always a list.
        if isinstance(source, string_types):
            sources = [source]
        elif isinstance(source, Iterable):
            sources = list(source)
        elif source:
            sources = [source]
        else:
            raise TopologyError('{} component "{}" requires at least one source'
                                .format(self.component_cls.__name__, self.name))

        if not Grouping.valid(group_on):
            raise TypeError('Invalid group_on given: {}'.format(group_on))

        self.sources = sources
        self.group_on = group_on

    def resolve_dependencies(self, specifications):
        """Modifies the bolt's sources to be references to other Specification
        objects. Also ensures that group_on fields are provided by the
        source Specifications.

        :param specifications: all of the specification objects for this
                               topology.
        :type specifications: dict
        """
        for i, source in enumerate(self.sources):

            # Resolve sources that are not a specification object.
            if not isinstance(source, Specification):
                if source in specifications:
                    self.sources[i] = specifications[source]
                else:
                    raise TopologyError(
                        'Missing source named {} from {}'.format(source, self))

                # Ensure that source always references a specification object.
                source = self.sources[i]

            # Ensure that group on field names exist in the source.
            if isinstance(self.group_on, list):
                outputs = source.component_cls.streams

                for group in self.group_on:
                    if group not in outputs:
                        raise TopologyError(
                            'Field {} does not exist in source {} as an '
                            'output'.format(group, outputs))



import unittest
import logging
from streamparse.dsl.topology import Topology, Grouping, TopologyError
from streamparse.dsl.spec import Spec
from streamparse.storm.spout import Spout
from streamparse.storm.bolt import Bolt

log = logging.getLogger(__name__)


class WordSpout(Spout):
    streams = ["word"]


class WordCountBolt(Bolt):
    streams = ["word", "count"]


class TopologyTests(unittest.TestCase):
    def test_basic_spec(self):
        class WordCount(Topology):
            word_spout = Spec(
                WordSpout,
                parallelism=2,
            )
            word_bolt = Spec(
                WordCountBolt,
                source=word_spout,
                group_on=Grouping.fields("word"),
                parallelism=8,
            )

        self.assertEqual(len(WordCount.specs), 2)
        self.assertEqual(WordCount.word_bolt.sources[0], WordCount.word_spout)
        self.assertEqual(WordCount.word_bolt.group_on, ['word'])

    def test_bolt_before_source(self):
        class WordCount(Topology):
            word_bolt = Spec(
                WordCountBolt,
                source='word_spout',
            )
            word_spout = Spec(WordSpout)

        self.assertEqual(WordCount.word_bolt.sources[0], WordCount.word_spout)

    def test_invalid_spout_source(self):
        with self.assertRaises(TypeError):
            class WordCount(Topology):
                word_spout = Spec(
                    WordSpout,
                    source='unknown',
                )

    def test_invalid_bolt_source(self):
        with self.assertRaises(TopologyError):
            class WordCount(Topology):
                word_bolt = Spec(
                    WordCountBolt,
                )

    def test_invalid_parallelism(self):
        with self.assertRaises(ValueError):
            class WordCount(Topology):
                word_spout = Spec(
                    WordSpout,
                    parallelism=0,
                )

    def test_invalid_bolt_group_on(self):
        with self.assertRaises(TopologyError):
            class WordCount(Topology):
                word_spout = Spec(
                    WordSpout,
                )
                word_bolt = Spec(
                    WordCountBolt,
                    source=word_spout,
                    group_on=Grouping.fields("foo"),
                )

    def test_duplicate_name(self):
        with self.assertRaises(TopologyError):
            class WordCount(Topology):
                word = Spec(
                    WordSpout,
                )
                word_ = Spec(
                    WordCountBolt,
                    name='word',
                    source=word,
                )



# -*- coding: utf-8 -*-
"""
streamparse.debug

Remote debugger for Python-based Storm tasks, based on Pdb.

Inspired by http://snippets.dzone.com/posts/show/7248.
"""
from __future__ import print_function, unicode_literals

import errno
import os
import socket
import sys

from pdb import Pdb

__all__ = ['StormPdb', 'default_port', 'debugger', 'set_trace']

default_port = 7999

HOST = os.environ.get('STORM_PDB_HOST') or '127.0.0.1'
PORT = int(os.environ.get('STORM_PDB_PORT') or default_port)

#: Holds the currently active debugger.
_current = [None]

_frame = getattr(sys, '_getframe')

NO_AVAILABLE_PORT = """\
{self.ident}: Couldn't find an available port.

Please specify one using the STORM_PDB_PORT environment variable.
"""

BANNER = """\
{self.ident}: Please telnet into {self.host} {self.port}.

Type `exit` in session to continue.

{self.ident}: Waiting for client...
"""

SESSION_STARTED = '{self.ident}: Now in session with {self.remote_addr}.'
SESSION_ENDED = '{self.ident}: Session with {self.remote_addr} ended.'


class StormPdb(Pdb):
    me = 'Storm pdb'
    _prev_outs = None
    _sock = None

    def __init__(self, host=HOST, port=PORT,
                 port_search_limit=100, port_skew=+0, out=sys.stdout):
        self.active = True
        self.out = out

        self._prev_handles = sys.stdin, sys.stdout

        self._sock, this_port = self.get_avail_port(
            host, port, port_search_limit, port_skew,
        )
        self._sock.setblocking(1)
        self._sock.listen(1)
        self.ident = '{0}:{1}'.format(self.me, this_port)
        self.host = host
        self.port = this_port
        self.say(BANNER.format(self=self))

        self._client, address = self._sock.accept()
        self._client.setblocking(1)
        self.remote_addr = ':'.join(str(v) for v in address)
        self.say(SESSION_STARTED.format(self=self))
        self._handle = sys.stdin = sys.stdout = self._client.makefile('rw')
        Pdb.__init__(self, completekey='tab',
                     stdin=self._handle, stdout=self._handle)

    def get_avail_port(self, host, port, search_limit=100, skew=+0):
        try:
            _, skew = "foo-0".split('-')
            skew = int(skew)
        except ValueError:
            pass
        this_port = None
        for i in range(search_limit):
            _sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            this_port = port + skew + i
            try:
                _sock.bind((host, this_port))
            except socket.error as exc:
                if exc.errno in [errno.EADDRINUSE, errno.EINVAL]:
                    continue
                raise
            else:
                return _sock, this_port
        else:
            raise IOError(NO_AVAILABLE_PORT.format(self=self))

    def say(self, m):
        print(m, file=self.out)

    def _close_session(self):
        self.stdin, self.stdout = sys.stdin, sys.stdout = self._prev_handles
        self._handle.close()
        self._client.close()
        self._sock.close()
        self.active = False
        self.say(SESSION_ENDED.format(self=self))

    def do_continue(self, arg):
        self._close_session()
        self.set_continue()
        return 1
    do_c = do_cont = do_continue

    def do_quit(self, arg):
        self._close_session()
        self.set_quit()
        return 1
    do_q = do_exit = do_quit

    def set_trace(self, frame=None):
        if frame is None:
            frame = _frame().f_back
        Pdb.set_trace(self, frame)


    def set_quit(self):
        # this raises a BdbQuit exception that we are unable to catch.
        sys.settrace(None)


def debugger():
    """Return the current debugger instance (if any),
    or creates a new one."""
    rdb = _current[0]
    if rdb is None or not rdb.active:
        rdb = _current[0] = StormPdb()
    return rdb


def set_trace(frame=None):
    """Set breakpoint at current location, or a specified frame"""
    if frame is None:
        frame = _frame().f_back
    return debugger().set_trace(frame)

if __name__ == "__main__":
    foo = 42
    set_trace()



"""
Component-level Specification

This module is called component to mirror organization of storm package.
"""

from ..storm.component import Component


class Specification(object):
    def __init__(self, component_cls, name=None, parallelism=1):
        if not issubclass(component_cls, Component):
            raise TypeError("Invalid component: {}".format(component_cls))

        if not isinstance(parallelism, int) or parallelism < 1:
            raise ValueError("Parallelism must be a integer greater than 0")

        self.component_cls = component_cls
        self.name = name
        self.parallelism = parallelism

    def resolve_dependencies(self, specifications):
        """Allows specification subclasses to resolve an dependencies
        that they may have on other specifications.

        :param specifications: all of the specification objects for this
                               topology.
        :type specifications: dict
        """
        pass


from .spout import SpoutSpecification
from .bolt import BoltSpecification
from ..storm.bolt import Bolt
from ..storm.spout import Spout


def Spec(component_cls, *args, **kwargs):
    """A Specification factory that allows Topologies to be defined via
    Spec(Specification, ...)
    """
    if issubclass(component_cls, Spout):
        return SpoutSpecification(component_cls, *args, **kwargs)
    elif issubclass(component_cls, Bolt):
        return BoltSpecification(component_cls, *args, **kwargs)
    else:
        raise TypeError("Unhandled component: {}".format(component_cls))



"""
Tests for basic IPC stuff via Component class
"""

from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import unittest
from io import BytesIO

import simplejson as json

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from streamparse.storm import Component, Tuple


log = logging.getLogger(__name__)


class ComponentTests(unittest.TestCase):

    def test_read_handshake(self):
        handshake_dict = {
            "conf": {
                "topology.message.timeout.secs": 3,
                "topology.tick.tuple.freq.secs": 1,
                "topology.debug": True
            },
            "pidDir": ".",
            "context": {
                "task->component": {
                    "1": "example-spout",
                    "2": "__acker",
                    "3": "example-bolt1",
                    "4": "example-bolt2"
                },
                "taskid": 3,
                # Everything below this line is only available in Storm 0.10.0+
                "componentid": "example-bolt1",
                "stream->target->grouping": {
                    "default": {
                        "example-bolt2": {
                            "type": "SHUFFLE"
                        }
                    }
                },
                "streams": ["default"],
                "stream->outputfields": {"default": ["word"]},
                "source->stream->grouping": {
                    "example-spout": {
                        "default": {
                            "type": "FIELDS",
                            "fields": ["word"]
                        }
                    }
                },
                "source->stream->fields": {
                    "example-spout": {
                        "default": ["word"]
                    }
                }
            }
        }
        pid_dir = handshake_dict['pidDir']
        expected_conf = handshake_dict['conf']
        expected_context = handshake_dict['context']
        inputs = ["{}\n".format(json.dumps(handshake_dict)),
                  "end\n"]
        component = Component(input_stream=BytesIO(''.join(inputs).encode('utf-8')),
                              output_stream=BytesIO())
        given_conf, given_context = component.read_handshake()
        pid_path = os.path.join(pid_dir, str(component.pid))
        self.assertTrue(os.path.exists(pid_path))
        os.remove(pid_path)
        self.assertEqual(given_conf, expected_conf)
        self.assertEqual(given_context, expected_context)
        self.assertEqual("{}\nend\n".format(json.dumps({"pid": component.pid})).encode('utf-8'),
                         component.output_stream.buffer.getvalue())

    def test_setup_component(self):
        conf = {"topology.message.timeout.secs": 3,
                "topology.tick.tuple.freq.secs": 1,
                "topology.debug": True,
                "topology.name": "foo"}
        context = {
            "task->component": {
                "1": "example-spout",
                "2": "__acker",
                "3": "example-bolt1",
                "4": "example-bolt2"
            },
            "taskid": 3,
            # Everything below this line is only available in Storm 0.11.0+
            "componentid": "example-bolt1",
            "stream->target->grouping": {
                "default": {
                    "example-bolt2": {
                        "type": "SHUFFLE"
                    }
                }
            },
            "streams": ["default"],
            "stream->outputfields": {"default": ["word"]},
            "source->stream->grouping": {
                "example-spout": {
                    "default": {
                        "type": "FIELDS",
                        "fields": ["word"]
                    }
                }
            },
            "source->stream->fields": {
                "example-spout": {
                    "default": ["word"]
                }
            }
        }
        component = Component(input_stream=BytesIO(),
                              output_stream=BytesIO())
        component._setup_component(conf, context)
        self.assertEqual(component.topology_name, conf['topology.name'])
        self.assertEqual(component.task_id, context['taskid'])
        self.assertEqual(component.component_name,
                         context['task->component'][str(context['taskid'])])
        self.assertEqual(component.storm_conf, conf)
        self.assertEqual(component.context, context)

    def test_read_message(self):
        inputs = [# Task IDs
                  '[12, 22, 24]\n', 'end\n',
                  # Incoming Tuple for bolt
                  ('{ "id": "-6955786537413359385", "comp": "1", "stream": "1"'
                   ', "task": 9, "tuple": ["snow white and the seven dwarfs", '
                   '"field2", 3]}\n'), 'end\n',
                  # next command for spout
                  '{"command": "next"}\n', 'end\n',
                  # empty message, which should trigger sys.exit (end ignored)
                  '', '']
        outputs = [json.loads(msg) for msg in inputs[::2] if msg]
        outputs.append('')
        component = Component(input_stream=BytesIO(''.join(inputs).encode('utf-8')),
                              output_stream=BytesIO())
        for output in outputs:
            log.info('Checking msg for %s', output)
            if output:
                msg = component.read_message()
                self.assertEqual(output, msg)
            else:
                with self.assertRaises(SystemExit):
                    component.read_message()

    def test_read_message_unicode(self):
        inputs = [# Task IDs
                  '[12, 22, 24]\n', 'end\n',
                  # Incoming Tuple for bolt
                  ('{ "id": "-6955786537413359385", "comp": "1", "stream": "1"'
                   ', "task": 9, "tuple": ["snow white \uFFE6 the seven dwarfs"'
                   ', "field2", 3]}\n'), 'end\n',
                  # next command for spout
                  '{"command": "next"}\n', 'end\n',
                  # empty message, which should trigger sys.exit (end ignored)
                  '', '']
        outputs = [json.loads(msg) for msg in inputs[::2] if msg]
        outputs.append('')
        component = Component(input_stream=BytesIO(''.join(inputs).encode('utf8')),
                              output_stream=BytesIO())
        for output in outputs:
            log.info('Checking msg for %s', output)
            if output:
                msg = component.read_message()
                self.assertEqual(output, msg)
            else:
                with self.assertRaises(SystemExit):
                    component.read_message()

    def test_read_split_message(self):
        # Make sure we can read something that's broken up into many "lines"
        inputs = ['{ "id": "-6955786537413359385", ',
                  '"comp": "1", "stream": "1"\n',
                  '\n',
                  ', "task": 9, "tuple": ["snow white and the seven dwarfs", ',
                  '"field2", 3]}\n',
                  'end\n']
        output = json.loads(''.join(inputs[:-1]))

        component = Component(input_stream=BytesIO(''.join(inputs).encode('utf-8')),
                              output_stream=BytesIO())
        msg = component.read_message()
        self.assertEqual(output, msg)

    def test_read_command(self):
        # Check that we properly queue task IDs and return only commands
        inputs = [# Task IDs
                  '[12, 22, 24]\n', 'end\n',
                  # Incoming Tuple for bolt
                  ('{ "id": "-6955786537413359385", "comp": "1", "stream": "1"'
                   ', "task": 9, "tuple": ["snow white and the seven dwarfs", '
                   '"field2", 3]}\n'), 'end\n',
                  # next command for spout
                  '{"command": "next"}\n', 'end\n']
        outputs = [json.loads(msg) for msg in inputs[::2]]
        component = Component(input_stream=BytesIO(''.join(inputs).encode('utf-8')),
                              output_stream=BytesIO())

        # Skip first output, because it's a task ID, and won't be returned by
        # read_command
        for output in outputs[1:]:
            log.info('Checking msg for %s', output)
            msg = component.read_command()
            self.assertEqual(output, msg)
        self.assertEqual(component._pending_task_ids.pop(), outputs[0])

    def test_read_task_ids(self):
        # Check that we properly queue commands and return only task IDs
        inputs = [# Task IDs
                  '[4, 8, 15]\n', 'end\n',
                  # Incoming Tuple for bolt
                  ('{ "id": "-6955786537413359385", "comp": "1", "stream": "1"'
                   ', "task": 9, "tuple": ["snow white and the seven dwarfs", '
                   '"field2", 3]}\n'), 'end\n',
                  # next command for spout
                  '{"command": "next"}\n', 'end\n',
                  # Task IDs
                  '[16, 23, 42]\n', 'end\n']
        outputs = [json.loads(msg) for msg in inputs[::2]]
        component = Component(input_stream=BytesIO(''.join(inputs).encode('utf-8')),
                              output_stream=BytesIO())

        # Skip middle outputs, because they're commands and won't be returned by
        # read_task_ids
        for output in (outputs[0], outputs[-1]):
            log.info('Checking msg for %s', output)
            msg = component.read_task_ids()
            self.assertEqual(output, msg)
        for output in outputs[1:-1]:
            self.assertEqual(component._pending_commands.popleft(), output)

    def test_read_tuple(self):
        # This is only valid for bolts, so we only need to test with task IDs
        # and Tuples
        inputs = [# Tuple with all values
                  ('{ "id": "-6955786537413359385", "comp": "1", "stream": "1"'
                   ', "task": 9, "tuple": ["snow white and the seven dwarfs", '
                   '"field2", 3]}\n'), 'end\n',
                  # Tick Tuple
                  ('{ "id": null, "task": -1, "comp": "__system", "stream": '
                   '"__tick", "tuple": [50]}\n'), 'end\n',
                  # Heartbeat Tuple
                  ('{ "id": null, "task": -1, "comp": "__system", "stream": '
                   '"__heartbeat", "tuple": []}\n'), 'end\n',
                  ]
        outputs = []
        for msg in inputs[::2]:
            output = json.loads(msg)
            output['component'] = output['comp']
            output['values'] = output['tuple']
            del output['comp']
            del output['tuple']
            outputs.append(Tuple(**output))

        component = Component(input_stream=BytesIO(''.join(inputs).encode('utf-8')),
                              output_stream=BytesIO())

        for output in outputs:
            log.info('Checking Tuple for %s', output)
            tup = component.read_tuple()
            self.assertEqual(output, tup)

    def test_send_message(self):
        component = Component(input_stream=BytesIO(), output_stream=BytesIO())
        inputs = [{"command": "emit", "id": 4, "stream": "", "task": 9,
                   "tuple": ["field1", 2, 3]},
                  {"command": "log", "msg": "I am a robot monkey."},
                  {"command": "next"},
                  {"command": "sync"}]
        for cmd in inputs:
            component.output_stream.close()
            component.output_stream = component._wrap_stream(BytesIO())
            component.send_message(cmd)
            self.assertEqual("{}\nend\n".format(json.dumps(cmd)).encode('utf-8'),
                             component.output_stream.buffer.getvalue())

        # Check that we properly skip over invalid input
        self.assertIsNone(component.send_message(['foo', 'bar']))

    def test_send_message_unicode(self):
        component = Component(input_stream=BytesIO(), output_stream=BytesIO())
        inputs = [{"command": "emit", "id": 4, "stream": "", "task": 9,
                   "tuple": ["field\uFFE6", 2, 3]},
                  {"command": "log", "msg": "I am a robot monkey."},
                  {"command": "next"},
                  {"command": "sync"}]
        for cmd in inputs:
            component.output_stream.close()
            component.output_stream = component._wrap_stream(BytesIO())
            component.send_message(cmd)
            self.assertEqual("{}\nend\n".format(json.dumps(cmd)).encode('utf-8'),
                             component.output_stream.buffer.getvalue())

        # Check that we properly skip over invalid input
        self.assertIsNone(component.send_message(['foo', 'bar']))

    @patch.object(Component, 'send_message', autospec=True)
    def test_log(self, send_message_mock):
        component = Component(input_stream=BytesIO(), output_stream=BytesIO())
        inputs = [("I am a robot monkey.", None, 2),
                  ("I am a monkey who learned to talk.", 'warning', 3)]
        for msg, level, storm_level in inputs:
            component.output_stream.close()
            component.output_stream = component._wrap_stream(BytesIO())
            component.log(msg, level=level)
            send_message_mock.assert_called_with(component, {'command': 'log',
                                                             'msg': msg,
                                                             'level': storm_level})


if __name__ == '__main__':
    unittest.main()



"""
Submit a Storm topology to Nimbus.
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
import re
import sys
import time

from invoke import run
from six import string_types

from .common import (add_ackers, add_debug, add_environment, add_name,
                     add_options, add_par, add_simple_jar, add_wait,
                     add_workers, resolve_ackers_workers)
from .jar import jar_for_deploy
from .kill import _kill_topology
from .list import _list_topologies
from .update_virtualenv import create_or_update_virtualenvs
from ..contextmanagers import ssh_tunnel
from ..util import (activate_env, get_config, get_env_config,
                    get_nimbus_for_env_config, get_topology_definition,
                    is_ssh_for_nimbus)


def get_user_tasks():
    """Get tasks defined in a user's tasks.py and fabric.py file which is
    assumed to be in the current working directory.

    :returns: a `tuple` (invoke_tasks, fabric_tasks)
    """
    sys.path.insert(0, os.getcwd())
    try:
        import tasks as user_invoke
    except ImportError:
        user_invoke = None
    try:
        import fabfile as user_fabric
    except ImportError:
        user_fabric = None
    return user_invoke, user_fabric


def is_safe_to_submit(topology_name, host=None, port=None):
    """Check to see if a topology is currently running or is in the process of
    being killed. Assumes tunnel is already connected to Nimbus."""
    result = _list_topologies(run_kwargs={"hide": "both"},
                              host=host, port=port)

    if result.failed:
        raise RuntimeError("Error running streamparse.commands.list/-main")

    pattern = re.compile(r"{}\s+\|\s+(ACTIVE|KILLED)\s+\|"
                         .format(topology_name))
    if re.search(pattern, result.stdout):
        return False
    else:
        return True


def _kill_existing_topology(topology_name, force, wait, host=None, port=None):
    if force and not is_safe_to_submit(topology_name, host=host, port=port):
        print("Killing current \"{}\" topology.".format(topology_name))
        sys.stdout.flush()
        _kill_topology(topology_name, run_kwargs={"hide": "both"},
                       wait=wait, host=host, port=port)
        while not is_safe_to_submit(topology_name, host=host, port=port):
            print("Waiting for topology {} to quit...".format(topology_name))
            sys.stdout.flush()
            time.sleep(0.5)
        print("Killed.")
        sys.stdout.flush()


def _submit_topology(topology_name, topology_file, topology_jar,
                     env_config, workers, ackers, options=None, debug=False,
                     host=None, port=None):
    jvm_opts = [
        "-Dstorm.jar={}".format(topology_jar),
        "-Dstorm.options=",
        "-Dstorm.conf.file=",
    ]
    os.environ["JVM_OPTS"] = " ".join(jvm_opts)
    cmd = [
        "lein",
        "run -m streamparse.commands.submit_topology/-main",
        topology_file]

    if host:
        cmd.append("--host {}".format(host))
    if port:
        cmd.append("--port {}".format(port))
    if debug:
        cmd.append("--debug")

    cmd.append("--option 'topology.workers={}'".format(workers))
    cmd.append("--option 'topology.acker.executors={}'".format(ackers))

    if env_config.get('use_virtualenv', True):
        python_path = '/'.join([env_config["virtualenv_root"],
                                topology_name, "bin", "python"])

        cmd.append("--option 'topology.python.path=\"{}\"'".format(python_path))

    # Python logging settings
    log_config = env_config.get("log", {})
    log_path = log_config.get("path") or env_config.get("log_path")
    print("Routing Python logging to {}.".format(log_path))
    sys.stdout.flush()
    if log_path:
        cmd.append("--option 'streamparse.log.path=\"{}\"'"
                   .format(log_path))
    if isinstance(log_config.get("max_bytes"), int):
        cmd.append("--option 'streamparse.log.max_bytes={}'"
                   .format(log_config["max_bytes"]))
    if isinstance(log_config.get("backup_count"), int):
        cmd.append("--option 'streamparse.log.backup_count={}'"
                   .format(log_config["backup_count"]))
    if isinstance(log_config.get("level"), string_types):
        cmd.append("--option 'streamparse.log.level=\"{}\"'"
                   .format(log_config["level"].lower()))

    if options is None:
        options = []
    for option in options:
        # XXX: hacky Parse.ly-related workaround; must fix root
        # issue with -o options and string values
        if "deployment_stage" in option:
            key, val = option.split("=")
            cmd.append("--option '{}=\"{}\"'".format(key, val))
        else:
            cmd.append("--option {}".format(option))
    full_cmd = " ".join(cmd)
    print("Running lein command to submit topology to nimbus:")
    print(full_cmd)
    sys.stdout.flush()
    run(full_cmd)


def _pre_submit_hooks(topology_name, env_name, env_config):
    """Pre-submit hooks for invoke and fabric.
    """
    user_invoke, user_fabric = get_user_tasks()
    pre_submit_invoke = getattr(user_invoke, "pre_submit", None)
    if callable(pre_submit_invoke):
        pre_submit_invoke(topology_name, env_name, env_config)
    pre_submit_fabric = getattr(user_fabric, "pre_submit", None)
    if callable(pre_submit_fabric):
        pre_submit_fabric(topology_name, env_name, env_config)


def _post_submit_hooks(topology_name, env_name, env_config):
    """Post-submit hooks for invoke and fabric.
    """
    user_invoke, user_fabric = get_user_tasks()
    post_submit_invoke = getattr(user_invoke, "post_submit", None)
    if callable(post_submit_invoke):
        post_submit_invoke(topology_name, env_name, env_config)
    post_submit_fabric = getattr(user_fabric, "post_submit", None)
    if callable(post_submit_fabric):
        post_submit_fabric(topology_name, env_name, env_config)


def submit_topology(name=None, env_name="prod", workers=2, ackers=2,
                    options=None, force=False, debug=False, wait=None,
                    simple_jar=False):
    """Submit a topology to a remote Storm cluster."""
    config = get_config()
    name, topology_file = get_topology_definition(name)
    env_name, env_config = get_env_config(env_name)
    host, port = get_nimbus_for_env_config(env_config)

    # Check if we need to maintain virtualenv during the process
    use_venv = env_config.get('use_virtualenv', True)
    # Setup the fabric env dictionary
    activate_env(env_name)
    # Run pre_submit actions provided by project
    _pre_submit_hooks(name, env_name, env_config)

    if use_venv:
        config["virtualenv_specs"] = config["virtualenv_specs"].rstrip("/")
        create_or_update_virtualenvs(
            env_name,
            name,
            "{}/{}.txt".format(config["virtualenv_specs"], name),
            virtualenv_flags=env_config.get('virtualenv_flags'))

    # Prepare a JAR that doesn't have Storm dependencies packaged
    topology_jar = jar_for_deploy(simple_jar=simple_jar)

    print('Deploying "{}" topology...'.format(name))
    sys.stdout.flush()
    # Use ssh tunnel with Nimbus or use host/port for Thrift connection
    if is_ssh_for_nimbus(env_config):
        with ssh_tunnel(env_config.get("user"), host, 6627, port):
            print("ssh tunnel to Nimbus {}:{} established.".format(host, port))
            sys.stdout.flush()
            _kill_existing_topology(name, force, wait)
            _submit_topology(name, topology_file, topology_jar,
                             env_config, workers, ackers, options, debug)
            _post_submit_hooks(name, env_name, env_config)
    else:
        # This part doesn't use SSH tunnel at all
        _kill_existing_topology(name, force, wait, host=host, port=port)
        _submit_topology(name, topology_file, topology_jar,
                         env_config, workers, ackers, options, debug,
                         host=host, port=port)
        _post_submit_hooks(name, env_name, env_config)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('submit',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_ackers(subparser)
    add_debug(subparser)
    add_environment(subparser)
    subparser.add_argument('-f', '--force',
                           action='store_true',
                           help='Force a topology to submit by killing any '
                                'currently running topologies with the same '
                                'name.')
    add_name(subparser)
    add_options(subparser)
    add_par(subparser)
    add_simple_jar(subparser)
    subparser.add_argument('-t', '--time',
                           default=0,
                           type=int,
                           help='Time (in seconds) to keep local cluster '
                                'running. If time <= 0, run indefinitely. '
                                '(default: %(default)s)')
    add_wait(subparser)
    add_workers(subparser)


def main(args):
    """ Submit a Storm topology to Nimbus. """
    resolve_ackers_workers(args)
    submit_topology(name=args.name, env_name=args.environment,
                    workers=args.workers, ackers=args.ackers,
                    options=args.options, force=args.force, debug=args.debug,
                    wait=args.wait, simple_jar=args.simple_jar)



"""Base bolt classes."""
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import signal
import sys
import threading
import time
from collections import defaultdict

from six import iteritems, itervalues, reraise

from .component import Component, Tuple


log = logging.getLogger(__name__)


class Bolt(Component):
    """The base class for all streamparse bolts.

    For more information on bolts, consult Storm's
    `Concepts documentation <http://storm.apache.org/documentation/Concepts.html>`_.

    :ivar auto_anchor: A ``bool`` indicating whether or not the bolt should
                       automatically anchor emits to the incoming Tuple ID.
                       Tuple anchoring is how Storm provides reliability, you
                       can read more about
                       `Tuple anchoring in Storm's docs <https://storm.apache.org/documentation/Guaranteeing-message-processing.html#what-is-storms-reliability-api>`_.
                       Default is ``True``.

    :ivar auto_ack: A ``bool`` indicating whether or not the bolt should
                    automatically acknowledge Tuples after ``process()``
                    is called. Default is ``True``.
    :ivar auto_fail: A ``bool`` indicating whether or not the bolt should
                     automatically fail Tuples when an exception occurs when the
                     ``process()`` method is called. Default is ``True``.

    **Example**:

    .. code-block:: python

        from streamparse.bolt import Bolt

        class SentenceSplitterBolt(Bolt):

            def process(self, tup):
                sentence = tup.values[0]
                for word in sentence.split(" "):
                    self.emit([word])
    """

    auto_anchor = True
    auto_ack = True
    auto_fail = True

    # Using list; Bolt class and subclasses can have more than one current_tup.
    _current_tups = []

    @staticmethod
    def is_tick(tup):
        """ :returns: Whether or not the given Tuple is a tick Tuple """
        return tup.component == '__system' and tup.stream == '__tick'

    def initialize(self, storm_conf, context):
        """Called immediately after the initial handshake with Storm and before
        the main run loop. A good place to initialize connections to data
        sources.

        :param storm_conf: the Storm configuration for this bolt.  This is the
                           configuration provided to the topology, merged in
                           with cluster configuration on the worker node.
        :type storm_conf: dict
        :param context: information about the component's place within the
                        topology such as: task IDs, inputs, outputs etc.
        :type context: dict
        """
        pass

    def process(self, tup):
        """Process a single Tuple :class:`streamparse.storm.component.Tuple` of
        input

        This should be overridden by subclasses.
        :class:`streamparse.storm.component.Tuple` objects contain metadata
        about which component, stream and task it came from. The actual values
        of the Tuple can be accessed by calling ``tup.values``.

        :param tup: the Tuple to be processed.
        :type tup: :class:`streamparse.storm.component.Tuple`
        """
        raise NotImplementedError()

    def process_tick(self, tup):
        """Process special 'tick Tuples' which allow time-based
        behaviour to be included in bolts.

        Default behaviour is to ignore time ticks.  This should be
        overridden by subclasses who wish to react to timer events
        via tick Tuples.

        Tick Tuples will be sent to all bolts in a toplogy when the
        storm configuration option 'topology.tick.tuple.freq.secs'
        is set to an integer value, the number of seconds.

        :param tup: the Tuple to be processed.
        :type tup: :class:`streamparse.storm.component.Tuple`
        """
        pass

    def emit(self, tup, stream=None, anchors=None, direct_task=None,
             need_task_ids=True):
        """Emit a new Tuple to a stream.

        :param tup: the Tuple payload to send to Storm, should contain only
                    JSON-serializable data.
        :type tup: :class:`list` or :class:`streamparse.storm.component.Tuple`
        :param stream: the ID of the stream to emit this Tuple to. Specify
                       ``None`` to emit to default stream.
        :type stream: str
        :param anchors: IDs the Tuples (or :class:`streamparse.storm.component.Tuple`
                        instances) which the emitted Tuples should be anchored
                        to. If ``auto_anchor`` is set to ``True`` and
                        you have not specified ``anchors``, ``anchors`` will be
                        set to the incoming/most recent Tuple ID(s).
        :type anchors: list
        :param direct_task: the task to send the Tuple to.
        :type direct_task: int
        :param need_task_ids: indicate whether or not you'd like the task IDs
                              the Tuple was emitted (default: ``True``).
        :type need_task_ids: bool

        :returns: a ``list`` of task IDs that the Tuple was sent to. Note that
                  when specifying direct_task, this will be equal to
                  ``[direct_task]``. If you specify ``need_task_ids=False``,
                  this function will return ``None``.
        """
        if anchors is None:
            anchors = self._current_tups if self.auto_anchor else []
        anchors = [a.id if isinstance(a, Tuple) else a for a in anchors]

        return super(Bolt, self).emit(tup, stream=stream, anchors=anchors,
                                      direct_task=direct_task,
                                      need_task_ids=need_task_ids)

    def emit_many(self, tuples, stream=None, anchors=None, direct_task=None,
                  need_task_ids=True):
        """Emit multiple Tuples.

        :param tuples: a ``list`` of multiple Tuple payloads to send to
                       Storm. All Tuples should contain only
                       JSON-serializable data.
        :type tuples: list
        :param stream: the ID of the steram to emit these Tuples to. Specify
                       ``None`` to emit to default stream.
        :type stream: str
        :param anchors: IDs the Tuples (or :class:`streamparse.storm.component.Tuple`
                        instances) which the emitted Tuples should be anchored
                        to. If ``auto_anchor`` is set to ``True`` and
                        you have not specified ``anchors``, ``anchors`` will be
                        set to the incoming/most recent Tuple ID(s).
        :type anchors: list
        :param direct_task: indicates the task to send the Tuple to.
        :type direct_task: int
        :param need_task_ids: indicate whether or not you'd like the task IDs
                              the Tuple was emitted (default:
                              ``True``).
        :type need_task_ids: bool

        .. deprecated:: 2.0.0
            Just call :py:meth:`Bolt.emit` repeatedly instead.
        """
        if not isinstance(tuples, (list, tuple)):
            raise TypeError('Tuples should be a list of lists/tuples, '
                            'received {!r} instead.'.format(type(tuples)))

        all_task_ids = []
        for tup in tuples:
            all_task_ids.append(self.emit(tup, stream=stream, anchors=anchors,
                                          direct_task=direct_task,
                                          need_task_ids=need_task_ids))

        return all_task_ids

    def ack(self, tup):
        """Indicate that processing of a Tuple has succeeded.

        :param tup: the Tuple to acknowledge.
        :type tup: :class:`str` or :class:`streamparse.storm.component.Tuple`
        """
        tup_id = tup.id if isinstance(tup, Tuple) else tup
        self.send_message({'command': 'ack', 'id': tup_id})

    def fail(self, tup):
        """Indicate that processing of a Tuple has failed.

        :param tup: the Tuple to fail (its ``id`` if ``str``).
        :type tup: :class:`str` or :class:`streamparse.storm.component.Tuple`
        """
        tup_id = tup.id if isinstance(tup, Tuple) else tup
        self.send_message({'command': 'fail', 'id': tup_id})

    def _run(self):
        """The inside of ``run``'s infinite loop.

        Separated out so it can be properly unit tested.
        """
        self._current_tups = [self.read_tuple()]
        try:
            tup = self._current_tups[0]
            if self.is_heartbeat(tup):
                self.send_message({'command': 'sync'})
            elif self.is_tick(tup):
                self.process_tick(tup)
                if self.auto_ack:
                     self.ack(tup)
            else:
                self.process(tup)
                if self.auto_ack:
                     self.ack(tup)
        finally:
            # reset so that we don't accidentally fail the wrong Tuples
            # if a successive call to read_tuple fails
            self._current_tups = []

    def _handle_run_exception(self, exc):
        """Process an exception encountered while running the ``run()`` loop.

        Called right before program exits.
        """
        super(Bolt, self)._handle_run_exception(exc)

        if len(self._current_tups) == 1:
            tup = self._current_tups[0]
            self.raise_exception(exc, tup)
            if self.auto_fail:
                self.fail(tup)


class BatchingBolt(Bolt):
    """A bolt which batches Tuples for processing.

    Batching Tuples is unexpectedly complex to do correctly. The main problem
    is that all bolts are single-threaded. The difficult comes when the
    topology is shutting down because Storm stops feeding the bolt Tuples. If
    the bolt is blocked waiting on stdin, then it can't process any waiting
    Tuples, or even ack ones that were asynchronously written to a data store.

    This bolt helps with that by grouping Tuples received between tick Tuples
    into batches.

    To use this class, you must implement ``process_batch``. ``group_key`` can
    be optionally implemented so that Tuples are grouped before
    ``process_batch`` is even called.

    You must also set the `topology.tick.tuple.freq.secs` to how frequently you
    would like ticks to be sent.  If you want ``ticks_between_batches`` to work
    the same way ``secs_between_batches`` worked in older versions of
    streamparse, just set `topology.tick.tuple.freq.secs` to 1.  This setting
    can be specified either at the topology level, or in the topology Clojure
    file by adding `:conf {"topology.tick.tuple.freq.secs", 1}` to your
    `python-bolt-spec`.


    :ivar auto_anchor: A ``bool`` indicating whether or not the bolt should
                       automatically anchor emits to the incoming Tuple ID.
                       Tuple anchoring is how Storm provides reliability, you
                       can read more about `Tuple anchoring in Storm's
                       docs <https://storm.apache.org/documentation/Guaranteeing-message-processing.html#what-is-storms-reliability-api>`_.
                       Default is ``True``.
    :ivar auto_ack: A ``bool`` indicating whether or not the bolt should
                    automatically acknowledge Tuples after ``process_batch()``
                    is called. Default is ``True``.
    :ivar auto_fail: A ``bool`` indicating whether or not the bolt should
                     automatically fail Tuples when an exception occurs when the
                     ``process_batch()`` method is called. Default is ``True``.
    :ivar ticks_between_batches: The number of tick Tuples to wait before
                                 processing a batch.


    **Example**:

    .. code-block:: python

        from streamparse.bolt import BatchingBolt

        class WordCounterBolt(BatchingBolt):

            ticks_between_batches = 5

            def group_key(self, tup):
                word = tup.values[0]
                return word  # collect batches of words

            def process_batch(self, key, tups):
                # emit the count of words we had per 5s batch
                self.emit([key, len(tups)])
    """

    auto_anchor = True
    auto_ack = True
    auto_fail = True
    ticks_between_batches = 1

    def __init__(self, *args, **kwargs):
        super(BatchingBolt, self).__init__(*args, **kwargs)
        self._batches = defaultdict(list)
        self._tick_counter = 0

    def group_key(self, tup):
        """Return the group key used to group Tuples within a batch.

        By default, returns None, which put all Tuples in a single
        batch, effectively just time-based batching. Override this to create
        multiple batches based on a key.

        :param tup: the Tuple used to extract a group key
        :type tup: :class:`streamparse.storm.component.Tuple`
        :returns: Any ``hashable`` value.
        """
        return None

    def process_batch(self, key, tups):
        """Process a batch of Tuples. Should be overridden by subclasses.

        :param key: the group key for the list of batches.
        :type key: hashable
        :param tups: a `list` of :class:`streamparse.storm.component.Tuple` s
                     for the group.
        :type tups: list
        """
        raise NotImplementedError()

    def emit(self, tup, **kwargs):
        """Modified emit that will not return task IDs after emitting.

        See :class:`streamparse.storm.component.Bolt` for more information.

        :returns: ``None``.
        """
        kwargs['need_task_ids'] = False
        return super(BatchingBolt, self).emit(tup, **kwargs)

    def emit_many(self, tups, **kwargs):
        """Modified emit_many that will not return task IDs after emitting.

        See :class:`streamparse.storm.component.Bolt` for more information.

        :returns: ``None``.

        .. deprecated:: 2.0.0
            Just call :py:meth:`BatchingBolt.emit` repeatedly instead.

        """
        kwargs['need_task_ids'] = False
        return super(BatchingBolt, self).emit_many(tups, **kwargs)

    def process_tick(self, tick_tup):
        """Increment tick counter, and call ``process_batch`` for all current
        batches if tick counter exceeds ``ticks_between_batches``.

        See :class:`streamparse.storm.component.Bolt` for more information.

        .. warning::
            This method should **not** be overriden.  If you want to tweak
            how Tuples are grouped into batches, override ``group_key``.
        """
        self._tick_counter += 1
        # ACK tick Tuple immediately, since it's just responsible for counter
        self.ack(tick_tup)
        if self._tick_counter > self.ticks_between_batches and self._batches:
            self.process_batches()
            self._tick_counter = 0

    def process_batches(self):
        """Iterate through all batches, call process_batch on them, and ack.

        Separated out for the rare instances when we want to subclass
        BatchingBolt and customize what mechanism causes batches to be
        processed.
        """
        for key, batch in iteritems(self._batches):
            self._current_tups = batch
            self.process_batch(key, batch)
            if self.auto_ack:
                for tup in batch:
                    self.ack(tup)
            # Set current batch to [] so that we know it was acked if a
            # later batch raises an exception
            self._batches[key] = []
        self._batches = defaultdict(list)

    def process(self, tup):
        """Group non-tick Tuples into batches by ``group_key``.

        .. warning::
            This method should **not** be overriden.  If you want to tweak
            how Tuples are grouped into batches, override ``group_key``.
        """
        # Append latest Tuple to batches
        group_key = self.group_key(tup)
        self._batches[group_key].append(tup)

    def _run(self):
        """The inside of ``run``'s infinite loop.

        Separated out so it can be properly unit tested.
        """
        self._current_tups = [self.read_tuple()]
        tup = self._current_tups[0]
        if self.is_heartbeat(tup):
            self.send_message({'command': 'sync'})
        elif self.is_tick(tup):
            self.process_tick(tup)
        else:
            self.process(tup)
        # reset so that we don't accidentally fail the wrong Tuples
        # if a successive call to read_tuple fails
        self._current_tups = []

    def _handle_run_exception(self, exc):
        """Process an exception encountered while running the ``run()`` loop.

        Called right before program exits.
        """
        # Don't use super here, because Bolt class has own auto fail handling.
        Component._handle_run_exception(self, exc)
        self.raise_exception(exc, self._current_tups)

        if self.auto_fail:
            # Fail batches
            for batch in itervalues(self._batches):
                for tup in batch:
                    self.fail(tup)
            # Fail current tick Tuple if we have one
            for tup in self._current_tups:
                if self.is_tick(tup):
                    self.fail(tup)


class TicklessBatchingBolt(BatchingBolt):
    """A BatchingBolt which uses a timer thread instead of tick tuples.

    Batching tuples is unexpectedly complex to do correctly. The main problem
    is that all bolts are single-threaded. The difficult comes when the
    topology is shutting down because Storm stops feeding the bolt tuples. If
    the bolt is blocked waiting on stdin, then it can't process any waiting
    tuples, or even ack ones that were asynchronously written to a data store.

    This bolt helps with that grouping tuples based on a time interval and then
    processing them on a worker thread.

    To use this class, you must implement ``process_batch``. ``group_key`` can
    be optionally implemented so that tuples are grouped before
    ``process_batch`` is even called.


    :ivar auto_anchor: A ``bool`` indicating whether or not the bolt should
                       automatically anchor emits to the incoming tuple ID.
                       Tuple anchoring is how Storm provides reliability, you
                       can read more about `tuple anchoring in Storm's
                       docs <https://storm.incubator.apache.org/documentation/Guaranteeing-message-processing.html#what-is-storms-reliability-api>`_.
                       Default is ``True``.
    :ivar auto_ack: A ``bool`` indicating whether or not the bolt should
                    automatically acknowledge tuples after ``process_batch()``
                    is called. Default is ``True``.
    :ivar auto_fail: A ``bool`` indicating whether or not the bolt should
                     automatically fail tuples when an exception occurs when the
                     ``process_batch()`` method is called. Default is ``True``.
    :ivar secs_between_batches: The time (in seconds) between calls to
                                ``process_batch()``. Note that if there are no
                                tuples in any batch, the TicklessBatchingBolt will
                                continue to sleep.

                                .. note::
                                  Can be fractional to specify greater
                                  precision (e.g. 2.5).

    **Example**:

    .. code-block:: python

        from streamparse.bolt import TicklessBatchingBolt

        class WordCounterBolt(TicklessBatchingBolt):

            secs_between_batches = 5

            def group_key(self, tup):
                word = tup.values[0]
                return word  # collect batches of words

            def process_batch(self, key, tups):
                # emit the count of words we had per 5s batch
                self.emit([key, len(tups)])
    """

    auto_anchor = True
    auto_ack = True
    auto_fail = True
    secs_between_batches = 2

    def __init__(self, *args, **kwargs):
        super(TicklessBatchingBolt, self).__init__(*args, **kwargs)
        self.exc_info = None
        signal.signal(signal.SIGUSR1, self._handle_worker_exception)

        iname = self.__class__.__name__
        threading.current_thread().name = '{}:main-thread'.format(iname)
        self._batches = defaultdict(list)
        self._batch_lock = threading.Lock()
        self._batcher = threading.Thread(target=self._batch_entry)
        self._batcher.name = '{}:_batcher-thread'.format(iname)
        self._batcher.daemon = True
        self._batcher.start()

    def process_tick(self, tick_tup):
        """ Just ack tick tuples and ignore them. """
        self.ack(tick_tup)

    def _batch_entry_run(self):
        """The inside of ``_batch_entry``'s infinite loop.

        Separated out so it can be properly unit tested.
        """
        time.sleep(self.secs_between_batches)
        with self._batch_lock:
            self.process_batches()

    def _batch_entry(self):
        """Entry point for the batcher thread."""
        try:
            while True:
                self._batch_entry_run()
        except Exception:
            self.exc_info = sys.exc_info()
            os.kill(self.pid, signal.SIGUSR1)  # interrupt stdin waiting

    def _handle_worker_exception(self, signum, frame):
        """Handle an exception raised in the worker thread.

        Exceptions in the _batcher thread will send a SIGUSR1 to the main
        thread which we catch here, and then raise in the main thread.
        """
        reraise(*self.exc_info)


    def _run(self):
        """The inside of ``run``'s infinite loop.

        Separate from BatchingBolt's implementation because
        we need to be able to acquire the batch lock after
        reading the tuple.

        We can't acquire the lock before reading the tuple because if
        that hange (i.e. the topology is shutting down) the lock being
        acquired will freeze the rest of the bolt, which is precisely
        what this batcher seeks to avoid.
        """
        tup = self.read_tuple()
        with self._batch_lock:
            self._current_tups = [tup]
            if self.is_heartbeat(tup):
                self.send_message({'command': 'sync'})
            elif self.is_tick(tup):
                self.process_tick(tup)
            else:
                self.process(tup)
            # reset so that we don't accidentally fail the wrong Tuples
            # if a successive call to read_tuple fails
            self._current_tups = []



"""
Helper script for starting up bolts and spouts.
"""

import argparse
import importlib


def main():
    """main entry point for Python bolts and spouts"""
    parser = argparse.ArgumentParser(description='Run a bolt/spout class',
                                     epilog='This is internal to streamparse '
                                            'and is used to run spout and bolt '
                                            'classes via ``python -m '
                                            'streamparse.run <class name>``.')
    parser.add_argument('target_class', help='The bolt/spout class to start.')
    args = parser.parse_args()
    mod_name, cls_name = args.target_class.rsplit('.', 1)
    mod = importlib.import_module(mod_name)
    cls = getattr(mod, cls_name)
    cls().run()


if __name__ == '__main__':
    main()



"""
Tests for Spout class
"""

from __future__ import absolute_import, print_function, unicode_literals

import logging
import unittest
from io import BytesIO

try:
    from unittest import mock
    from unittest.mock import patch
except ImportError:
    import mock
    from mock import patch

from streamparse.storm import Spout, Tuple


log = logging.getLogger(__name__)


class SpoutTests(unittest.TestCase):

    def setUp(self):
        self.tup_dict = {'id': 14,
                         'comp': 'some_spout',
                         'stream': 'default',
                         'task': 'some_spout',
                         'tuple': [1, 2, 3]}
        self.tup = Tuple(self.tup_dict['id'], self.tup_dict['comp'],
                         self.tup_dict['stream'], self.tup_dict['task'],
                         self.tup_dict['tuple'],)
        self.spout = Spout(input_stream=BytesIO(),
                           output_stream=BytesIO())
        self.spout.initialize({}, {})
        self.spout.logger = log

    @patch.object(Spout, 'send_message', autospec=True)
    def test_emit(self, send_message_mock):
        # A basic emit
        self.spout.emit([1, 2, 3], need_task_ids=False)
        send_message_mock.assert_called_with(self.spout, {'command': 'emit',
                                                          'tuple': [1, 2, 3],
                                                          'need_task_ids': False})

        # Emit as a direct task
        self.spout.emit([1, 2, 3], direct_task='other_spout')
        send_message_mock.assert_called_with(self.spout, {'command': 'emit',
                                                          'tuple': [1, 2, 3],
                                                          'task': 'other_spout'})
        # Reliable emit
        self.spout.emit([1, 2, 3], tup_id='foo', need_task_ids=False)
        send_message_mock.assert_called_with(self.spout, {'command': 'emit',
                                                          'tuple': [1, 2, 3],
                                                          'need_task_ids': False,
                                                          'id': 'foo'})

        # Reliable emit as direct task
        self.spout.emit([1, 2, 3], tup_id='foo', direct_task='other_spout')
        send_message_mock.assert_called_with(self.spout, {'command': 'emit',
                                                          'tuple': [1, 2, 3],
                                                          'task': 'other_spout',
                                                          'id': 'foo'})


    @patch.object(Spout, 'send_message', autospec=True)
    def test_emit_many(self, send_message_mock):
        # A basic emit
        self.spout.emit_many([[1, 2, 3], [4, 5, 6]], need_task_ids=False)
        send_message_mock.assert_has_calls([mock.call(self.spout,
                                                      {'command': 'emit',
                                                       'tuple': [1, 2, 3],
                                                       'need_task_ids': False}),
                                            mock.call(self.spout,
                                                      {'command': 'emit',
                                                       'tuple': [4, 5, 6],
                                                       'need_task_ids': False})])

        # Reliable emit
        send_message_mock.reset_mock()
        self.spout.emit_many([[1, 2, 3], [4, 5, 6]], tup_ids=['foo', 'bar'],
                             need_task_ids=False)
        send_message_mock.assert_has_calls([mock.call(self.spout,
                                                      {'command': 'emit',
                                                       'tuple': [1, 2, 3],
                                                       'need_task_ids': False,
                                                       'id': 'foo'}),
                                            mock.call(self.spout,
                                                      {'command': 'emit',
                                                       'tuple': [4, 5, 6],
                                                       'need_task_ids': False,
                                                       'id': 'bar'})])

        # Emit as a direct task
        send_message_mock.reset_mock()
        self.spout.emit_many([[1, 2, 3], [4, 5, 6]], direct_task='other_spout')
        send_message_mock.assert_has_calls([mock.call(self.spout,
                                                      {'command': 'emit',
                                                       'tuple': [1, 2, 3],
                                                       'task': 'other_spout'}),
                                            mock.call(self.spout,
                                                      {'command': 'emit',
                                                       'tuple': [4, 5, 6],
                                                       'task': 'other_spout'})])

    @patch.object(Spout, 'read_command', autospec=True,
                  return_value={'command': 'ack', 'id': 1234})
    @patch.object(Spout, 'ack', autospec=True)
    def test_ack(self, ack_mock, read_command_mock):
        # Make sure ack gets called
        self.spout._run()
        read_command_mock.assert_called_with(self.spout)
        ack_mock.assert_called_with(self.spout, 1234)

    @patch.object(Spout, 'read_command', autospec=True,
                  return_value={'command': 'fail', 'id': 1234})
    @patch.object(Spout, 'fail', autospec=True)
    def test_fail(self, fail_mock, read_command_mock):
        # Make sure fail gets called
        self.spout._run()
        read_command_mock.assert_called_with(self.spout)
        fail_mock.assert_called_with(self.spout, 1234)

    @patch.object(Spout, 'read_command', autospec=True,
                  return_value={'command': 'next', 'id': 1234})
    @patch.object(Spout, 'next_tuple', autospec=True)
    def test_next_tuple(self, next_tuple_mock, read_command_mock):
        self.spout._run()
        read_command_mock.assert_called_with(self.spout)
        self.assertEqual(next_tuple_mock.call_count, 1)


if __name__ == '__main__':
    unittest.main()
